# common procedures 


# quote ' and \  (backslash)
proc myquote { in_string } {

	#string --> list of char
        set in_string_list [split "$in_string" {}]
        #puts "|$in_string_list|"

        set out_string ""

	#we copy each element of the character list into
	#the output string and we double it if it is ' or \ (backslash)
        foreach el $in_string_list {
                switch -exact -- $el {

                        '
                                {
                                append out_string $el
                                append out_string $el
                                }
                        \\
                                {
                        #ending a line by \ (backslash) means that we continue on the next line, 
			#so we have to 'quote' backslash case in order to prevent script parsing error 
                                append out_string $el
                                append out_string $el
                                }
                        default
                                {
                                append out_string $el
                                }

                        } ;# end of switch


                } ;# end of foreach

        return $out_string

        }





proc distrib { dbname }  {
        global pgrdVar

# extracts the table replication map defining:
# 	pgrdVar($dbname,lhost) ... local host name x dbname
#       pgrdVar($dbname,lsite) ... local site-id x dbname
#       pgrdVar($dbname,rhosts)... list of remote hosts  x dbname
#       pgrdVar($dbname,$host,tables)...list of replicated tables on host x dbname
#       pgrdvar($dbname,$host,tables2resolv)... list of replicated tables to resolv on host x dbname 

set cfg_err_msg "Replicated database $dbname is not properly configured.\nPlease read related documentation to properly configure Replica System Tables in $dbname"

#set lfout $pgrdVar(pgrdlog)
#	if {[info exist pgrdVar($dbname,lconn)]} {
#	        set lconn $pgrdVar($dbname,lconn)
#		} else {
#		db_connect $dbname localhost $pgrdVar(pg_port)
#		set lconn $pgrdVar($dbname,lconn)
#		}

	if {[catch {db_testConnect $dbname localhost $pgrdVar(pg_port)} lconn]} {
		return -code error "$lconn"
		}

#puts $lfout "[clock clicks] distrib in for $dbname"


	# check if dbname is a replicated database
	set isreplicated 0
	set query "select count(*) as isreplicated from pg_class where relname = \'pgr_tables\'"
	set qry_handle [query_exec $dbname $query -return_handle]
	pg_result $qry_handle -assign risp
	set isreplicated $risp(0,isreplicated)
	pg_result $qry_handle -clear
	if { $isreplicated == 0} {
	        db_disconnect $dbname
	        return -code error "$dbname is not a replicated database!"
	        }


        catch {unset pgrdVar($dbname,lhost) pgrdVar($dbname,lsite) pgrdVar($dbname,rhosts) pgrdVar($dbname,$host,tables) tables}
        set pgrdVar($dbname,rhosts) {}
        # set pgrdVar($dbname,lhost) and pgrdVar($dbname,rhosts) list
	pg_select $lconn "select pgr_sid from pgr_lsite" S {
		set pgrdVar($dbname,lsite) $S(pgr_sid)
		}
	if {![info exists pgrdVar($dbname,lsite)]} {
		puts stderr "$cfg_err_msg"
		return -code error "No value in pgr_lsite"
		}

        pg_select $lconn "select host,domain,site  from pgr_hosts" H {
                set tables($H(host)) {} ;# inizializza tables(host)
                if { $H(site) == $pgrdVar($dbname,lsite) } {
                        set pgrdVar($dbname,lhost) $H(host)
			set pgrdVar($dbname,lsite,domain) $H(domain)
                        } else {
                        lappend pgrdVar($dbname,rhosts) $H(host)
                        }
                }

        if {![info exists pgrdVar($dbname,lhost)]} {
		puts stderr "$cfg_err_msg"
                return -code error "local host not defined in pgr_hosts"
                }
        if {![info exists pgrdVar($dbname,rhosts)]} {
		puts stderr "$cfg_err_msg"
                return -code error "any remote host defined in pgr_hosts"
                }



        # defining tables(host) 
        pg_select $lconn "select  host, tablename  from pgr_hosts a, pgr_distrib b where  a.site=b.site" HT {
                lappend tables($HT(host)) $HT(tablename)
                #puts $lfout "$HT(host): $tables($HT(host))"
                }

	if { [llength $tables($pgrdVar($dbname,lhost))]==0 } {
                return -code error "no replicated tables in pgr_tables"
                }


        # moves tables($host) into pgrdVar($dbname,$host,tables)
        foreach host [array names tables] {
                set  pgrdVar($dbname,$host,tables) $tables($host)
                unset tables($host)
                }


        # defining tables2resolv(host) 
        pg_select $lconn "select  host, tablename  from pgr_hosts a, pgr_tables b where  a.site=b.resolv_site" HT {
                lappend tables($HT(host)) $HT(tablename)
                #puts $lfout "$HT(host): $tables($HT(host))"
                }


        # moving tables2resolv($host) into pgrdVar($dbname,$host,tables2resolv)
        foreach host [array names tables] {
                set  pgrdVar($dbname,$host,tables2resolv) $tables($host)
                unset tables($host)
                }

	# set null value for pgrdVar($dbname,$host,tables2resolv) for hosts that are not resolv sites
	set hosts [linsert $pgrdVar($dbname,rhosts) 0 $pgrdVar($dbname,lhost)]
	foreach host $hosts {
		if {![info exists pgrdVar($dbname,$host,tables2resolv)]} {
			set pgrdVar($dbname,$host,tables2resolv) {}
			}
		}
	

#puts $lfout "[clock clicks] distrib out for $dbname"
        }  ;# end of proc


