#///////////////////////////////////////////////#
#   Large Object Procedures			#
#						#
#						#
#   LO_vacuum {dbname} 				#
#   setLOpgrdVar { dbname } 			#
#   find_LO_list { dbname flowtable } 		#
#   LO_export { dbname } 			#
#   LO_import { dbname } 			#
#   LO_update { dbname } 			#
#   LO_update_back { dbname } 			#
#						#
#///////////////////////////////////////////////#

proc LO_vacuum {dbname} {
# delete from pg_largeobject orphan loids
# it is very similar to vacuumlo but works on
# type 'lo' without requiring cast functions into
# the database

global pgrdVar
set lfout $pgrdVar(pgrdlog)
set lhost $pgrdVar($dbname,lhost)
set ldbfout $pgrdVar($dbname,pgrdlog)
set lsite $pgrdVar($dbname,lsite)
set lconn $pgrdVar($dbname,lconn)
set lofout $pgrdVar($dbname,lofout)

set now [clock format [clock second] -format %T]

foreach elem [array names pgrdVar] {puts $ldbfout "$elem - $pgrdVar($elem)"}

if {![info exists pgrdVar(vacuumLO)]} {
	set pgrdVar(vacuumLO) 0
	}

if { $pgrdVar(vacuumLO) == 0 } {
	puts $ldbfout [format "$now [clock clicks] vacuumLO disabled"]
	puts $lofout [format "$now [clock clicks] vacuumLO disabled"]
	return
	}

puts $ldbfout [format "$now [clock clicks] LO_vacuum IN\t$dbname"]
puts $lofout [format "$now [clock clicks] LO_vacuum IN\t$dbname"]

set now "        "


# search for ALL user tables with lo type fields
set query "select c.relname as table, a.attname as lofield from pg_class c, pg_attribute a, pg_type t where a.attrelid=c.oid  and a.atttypid=t.oid and  t.typname=\'lo\' and a.attnum>0 and c.relname not like \'pg\\_%\'"
catch {unset listlofields}
pg_select $lconn $query lofields {
        lappend listlofields($lofields(table)) $lofields(lofield)
	}

# drop/create vacuum_lo table
set query "drop table pgr_vacuumlo"
puts $lofout  "$now [clock clicks] $query..."
if {[catch {query_exec $dbname $query} ret]} {
        puts $lofout  "$now [clock clicks] ...$ret"
        } else {
        puts $lofout  "$now [clock clicks] ...$ret"
        }


# populating pgr_loids with distinct pg_largeobject loid 
set query "select distinct loid into pgr_vacuumlo from pg_largeobject"
puts $lofout  "$now [clock clicks] $query..."
if {[catch {query_exec $dbname $query} ret]} {
        puts $ldbfout  "$now [clock clicks] ...$ret"
        puts $lofout  "$now [clock clicks] ...$ret"
	return -code error "$ret"
        } else {
        puts $lofout  "$now [clock clicks] ...$ret"
        }


# creating index on pgr_vacuumlo
puts -nonewline $lofout "$now [clock clicks] creating index on pgr_vacuumlo..."
query_exec $dbname "create index pgr_vacuumlo_idx on pgr_vacuumlo (loid)"
puts $lofout " Done"


# scan tables with lo fields deleting LO value from pgr_vacuumlo
# so at the end of this process pgr_vacuumlo will get 
# the orphan loids
puts $ldbfout  "$now [clock clicks] scan tables with LO fields deleting LO value from pgr_vacuumlo..."
foreach table [array names listlofields] {

	# table lofields
	set lofields  $listlofields($table)
	set lofieldsString [join $lofields ", "]
	puts $lofout "$now [clock clicks] working on $table: $lofieldsString"
	
	set query "select $lofieldsString from $table"
	pg_select $lconn $query row {
		foreach lofield $lofields {
			if {[string compare $row($lofield) ""]!=0} {
				query_exec $dbname "delete from pgr_vacuumlo where loid=$row($lofield)"
				}
			};# end of foreach

		};# end of pg_select 

	};# end of foreach



# deleting from pg_largeobject where loid in pgr_vacuumlo
set query "delete from pg_largeobject where loid in (select loid from pgr_vacuumlo)"
puts $lofout  "$now [clock clicks] $query..."
if {[catch {query_exec $dbname $query} ret]} {
        puts $ldbfout  "$now [clock clicks] ...$ret"
        puts $lofout  "$now [clock clicks] ...$ret"
        return -code error "$ret"
        } else {
        puts $lofout  "$now [clock clicks] ...$ret"
        }


# vacuuming pg_largeobject
set query "vacuum pg_largeobject"
puts $lofout  "$now [clock clicks] $query..."
set ris [query_exec $dbname $query]
puts $lofout  "$now [clock clicks] $ris"

set now [clock format [clock second] -format %T]
puts $ldbfout [format "$now [clock clicks] LO_vacuum OUT\t$dbname"]
puts $lofout [format "$now [clock clicks] LO_vacuum OUT\t$dbname"]

}



#/////////////////////////////////////////////////
#/////////////////////////////////////////////////

proc setLOpgrdVar { dbname } {
global pgrdVar
set lfout $pgrdVar(pgrdlog)
set lhost $pgrdVar($dbname,lhost)
set ldbfout $pgrdVar($dbname,pgrdlog)
set lsite $pgrdVar($dbname,lsite)
set lconn $pgrdVar($dbname,lconn)
set lofout $pgrdVar($dbname,lofout)

# query to FIND LO fields in flow_tables
# Ex:
# listlofields(tab1_flow) = lofield1
# listlofields(tab2_flow) = lofield2 lofield3
set query "select c.relname as flowtable, a.attname as lofield from pg_class c, pg_attribute a, pg_type t where a.attrelid=c.oid  and a.atttypid=t.oid and  t.typname=\'lo\' and a.attnum>0 and c.relname not like \'pg\\_%\' and c.relname like \'%\\_flow\'"

puts $lofout "FIND LO fields in flow_tables"
puts $lofout "$query"
#puts $ldbfout "query: $query"

catch {unset listlofields}
pg_select $lconn $query lofields {
        lappend listlofields($lofields(flowtable)) $lofields(lofield)
}

# list of flow_tables having 1 or more 'lo' type fields
# Ex:
# listlotables = tab1_flow tab2_flow
set listlotables [array names listlofields]
set pgrdVar($dbname,lotables) $listlotables
foreach flowtable $listlotables {
        puts  $lofout "$flowtable: $listlofields($flowtable)"
        set pgrdVar($dbname,$flowtable,lofields)  $listlofields($flowtable)
	}

}

#/////////////////////////////////////////////////
#/////////////////////////////////////////////////

proc find_LO_list { dbname flowtable } {
global pgrdVar
set lfout $pgrdVar(pgrdlog)
set lhost $pgrdVar($dbname,lhost)
set ldbfout $pgrdVar($dbname,pgrdlog)
set lsite $pgrdVar($dbname,lsite)
set lconn $pgrdVar($dbname,lconn)
set lofout $pgrdVar($dbname,lofout)

set now [clock format [clock second] -format %T]
puts $ldbfout "$now [clock clicks] find_LO_list IN    $dbname $flowtable"
puts $lofout "$now [clock clicks] find_LO_list IN    $dbname $flowtable"

# flowtable lofields
set lofields  $pgrdVar($dbname,$flowtable,lofields)
set lofieldsString [join $lofields ", "]

# initialize loid list to be returned
set loidlist {}

set rec_state first_rec
set lop ""
catch {unset loidarray}
foreach lofield $lofields {
	set loidarray($lofield) ""
	}

set query "select pgr_grid, pgr_optype, $lofieldsString from  [myquote $flowtable] where  pgr_sid_f(pgr_gfid)=$lsite  and  pgr_flow_flag='\Y\' order by pgr_grid, pgr_gfid, pgr_optype desc"

puts $lofout $query
puts $lofout [format "\npgr_grid | pgr_optype | [join $lofields " | "] | rec_state | [join $lofields " | "]\n"]
# analizing flowtables rows
pg_select $lconn $query Trans {


	if {[string compare $rec_state first_rec]!=0} {
		if {$grid == $Trans(pgr_grid)} {
			set rec_state same_rec
			} else {
			set rec_state new_rec
			set grid $Trans(pgr_grid)
			}
		}

	switch -exact -- $rec_state {
        	first_rec {
			set grid $Trans(pgr_grid)
			foreach lofield $lofields {
				set loidarray($lofield) $Trans($lofield)
				}
			set rec_state new_rec
	                }
        	new_rec {
			foreach lofield $lofields {
				if {[string compare $loidarray($lofield) ""]!=0 && [string compare $lop "D"]!=0} {
					lappend loidlist $loidarray($lofield)
					}
				set loidarray($lofield) $Trans($lofield)
				}
	                }
        	same_rec {
			foreach lofield $lofields {
				if {[string compare $loidarray($lofield) $Trans($lofield)]==0} {
					set loidarray($lofield) ""
					} else {
					set loidarray($lofield) $Trans($lofield)
					}
				}
	                }
        	default {
	                puts $lofout "find_LO_list Error: $rec_state not in <first_rec | new_rec | same_rec>"
	                puts $ldbfout "find_LO_list Error: $rec_state not in <first_rec | new_rec | same_rec>"
                	return -1
	                }
        	} ;# end of switch

	set loglist {}
	lappend loglist $Trans(pgr_grid) $Trans(pgr_optype) 
	foreach lofield $lofields { lappend loglist $Trans($lofield) }
	lappend loglist $rec_state 
	foreach lofield $lofields { lappend loglist $loidarray($lofield) }
	set logString [join $loglist " | "]

	puts $lofout $logString 

	set lop $Trans(pgr_optype)

	} ;# end of pg_select 


# it was the last_rec:
foreach lofield $lofields {
	if {[string compare $loidarray($lofield) ""]!=0 && [string compare $lop "D"]!=0} {
		lappend loidlist $loidarray($lofield)
		}
	}


puts $lofout [format "\nloidlist: $loidlist"]


#sort lolist avoiding duplicate values 
puts $lofout  "$now [clock clicks] sort loidlist avoiding duplicate values"
set uniquesortlolist [lsort -integer -unique $loidlist]
puts  $lofout "uniquesortlolist: $uniquesortlolist"
puts  $lofout " "


set now [clock format [clock second] -format %T]
puts $ldbfout [format "$now [clock clicks] find_LO_list OUT\t$dbname $flowtable"]
puts $lofout [format "\n$now [clock clicks] find_LO_list OUT\t$dbname $flowtable\n"]

#return $loidlist
return $uniquesortlolist

} ;# end of proc find_LO_list




	
#/////////////////////////////////////////////////
#/////////////////////////////////////////////////

proc LO_export { dbname } {

global pgrdVar
set lfout $pgrdVar(pgrdlog)
set lhost $pgrdVar($dbname,lhost)
set ldbfout $pgrdVar($dbname,pgrdlog)
set lsite $pgrdVar($dbname,lsite)
set lconn $pgrdVar($dbname,lconn)


# Large Object log file (declared in RSTchk proc) 
#set logfname [file join $pgrdVar($dbname,log) LOprocs.log]
#set pgrdVar($dbname,lofout) [open $logfname w+]
#fconfigure $lofout -buffering none
set lofout $pgrdVar($dbname,lofout)


set now [clock format [clock second] -format %T]
puts $ldbfout [format "$now [clock clicks] Large Object Export IN\t$dbname"]
puts $lofout [format "$now [clock clicks] Large Object Export IN\t$dbname"]


set now "        "



#set curDir [pwd]
cd $pgrdVar($dbname,outbox)
#puts $lfout "sono in $pgrdVar($dbname,outbox)"



if {![info exists pgrdVar($dbname,lotables)]} {
	setLOpgrdVar $dbname
	}

# lista dei LO da tab_flow con oper. locali accettate
foreach flowtable $pgrdVar($dbname,lotables) {

	# find tablename from flowtablename
	regexp {(.+)_flow} $flowtable match table

	# find Lo-list
	puts $lofout  "$now [clock clicks] find_LO_list $dbname $flowtable..."
	#set lolist [find_LO_list $dbname $flowtable]
	set uniquesortlolist [find_LO_list $dbname $flowtable]

        # sort lolist avoiding duplicate values 
	#puts $lofout  "$now [clock clicks] sort lolist avoiding duplicate values"
        #set uniquesortlolist [lsort -integer -unique $lolist]
        #puts  $lofout "uniquesortlolist: $uniquesortlolist"
        #puts  $lofout " "


        # removing files.lo in  $pgrdVar($dbname,outbox)
	puts $lofout "$now [clock clicks] removing files.lo in  $pgrdVar($dbname,outbox)"
        set lo_list [split [glob -nocomplain -directory $pgrdVar($dbname,outbox) -- *.lo] { }]
        puts  $lofout "deleted: $lo_list"
        puts  $lofout " "
        foreach lofile $lo_list {
                file delete $lofile
        }


        # export table_lofiles
	puts $lofout "$now [clock clicks] export table_lofiles"
        set lofiles {}
        foreach lobjId $uniquesortlolist {
		set lofilename site${lsite}_${table}_${lobjId}.lo
                set abslofilename [file join $pgrdVar($dbname,outbox) $lofilename]
                pg_exec $lconn "BEGIN"
                # esporta il LO avente loid=lobjId in pg_largeobject in lofilename
                pg_lo_export $lconn $lobjId $abslofilename
                pg_exec $lconn "END"
                lappend lofiles $lofilename
        }
        puts  $lofout "exported: $lofiles"
        puts  $lofout " "


        # creating tgz file
	puts $lofout "$now [clock clicks] creating tgz file"
	catch { unset absarchive }
        if {[llength $lofiles]>0} {
                set archive ${lhost}_${table}_lofiles.tgz
                set absarchive [file join $pgrdVar($dbname,outbox)  $archive]
                puts   $lofout "tar cvzf  $archive  $lofiles..."
                if {[catch {eval exec tar cvzf $archive $lofiles} errmsg]} {
                        puts   $ldbfout "$now [clock clicks] ERROR: $errmsg"
                        puts   $lofout "$now [clock clicks] ERROR: $errmsg"
                        return -code error $errmsg
                } else {
                        puts   $lofout "...tar OK"
                }

	        # removing files.lo
        	set lo_list [split [glob -nocomplain -directory $pgrdVar($dbname,outbox) -- *.lo] { }]
	        puts $lofout "$now [clock clicks] removing files.lo in  $pgrdVar($dbname,outbox)"
	        puts   -nonewline $lofout "deleting: $lo_list ..."
	        foreach lofile $lo_list {
	                file delete $lofile
	        }
	        puts  $lofout "...Done"
			
	
	        # sending tgz file to remote_hosts(table)
	        # find rhosts for table
	        puts $lfout "$now [clock clicks] sending $absarchive  to $table remote hosts"
	        set query "select host from pgr_hosts where site in (select site from pgr_distrib where tablename=\'$table\') and site != $lsite;"
	        puts  $lofout "$query"
	        pg_select $lconn $query hostname {
	                set rhost $hostname(host)
	                # chiamo la procedura sendData
	                puts $lofout "[clock clicks] $absarchive --> $rhost ..."
	                set fname_send_done [sendData $dbname $absarchive $rhost]
	                puts $lofout "[clock clicks] ...$fname_send_done sent to $rhost"
	        }
	
	
	        # removing tgz file
		puts  -nonewline $lfout "$now [clock clicks] removing $absarchive..."
	        file delete $absarchive
		puts  $lfout " Done"

        } else {

                puts  $lofout "[clock clicks] No files.lo to export for $table"
                puts  $lofout " "

        }


} ;# end of 'foreach flowtable'


set now [clock format [clock second] -format %T]

puts $lofout [format "$now [clock clicks]  Large Object Export:   End of procedure\n"]
puts $ldbfout "$now [clock clicks]  Large Object Export:   End of procedure"


}





#/////////////////////////////////////////////////
#/////////////////////////////////////////////////

proc LO_import { dbname } {

global pgrdVar
set lfout $pgrdVar(pgrdlog)
set lhost $pgrdVar($dbname,lhost)
set ldbfout $pgrdVar($dbname,pgrdlog)
set lsite $pgrdVar($dbname,lsite)
set lconn $pgrdVar($dbname,lconn)
set lofout $pgrdVar($dbname,lofout)

set now [clock format [clock second] -format %T]
puts $ldbfout [format "$now [clock clicks] Large Object Import IN\t$dbname"]
puts $lofout [format "\n$now [clock clicks] Large Object Import IN\t$dbname"]
set now "        "

# rimuovo / ricreo pgr_lomap
set query "drop table pgr_lomap"
puts $lofout  "$now [clock clicks] $query..."
if {[catch {query_exec $dbname $query} ret]} {
	puts $lofout  "$now ...$ret"
	} else {
	puts $lofout  "$now ...$ret"
	}

set query "create table pgr_lomap (rsiteid int2, rloid int4, lloid int4)"
puts $lofout  "$now [clock clicks] $query..."
if {[catch {query_exec $dbname $query} ret]} {
        puts $lofout  "$now ...$ret"
        } else {
        puts $lofout  "$now ...$ret"
        }




#set curDir [pwd]

# cerco in $PGRBOX/dbname/in lofiles.tgz da scompattare 
# e importo i files in pg_largeobject
cd $pgrdVar($dbname,inbox)
set tgz_list [split [glob -nocomplain -- *lofiles.tgz] { }]
puts $lofout "$now [clock clicks] tgz_list: $tgz_list"
foreach tgz_file $tgz_list {

        # removing files.lo in  $pgrdVar($dbname,inbox)
        puts $ldbfout "removing files.lo in  $pgrdVar($dbname,inbox)"
        set lo_list [split [glob -nocomplain -directory $pgrdVar($dbname,inbox) -- *.lo] { }]
        puts  -nonewline $lofout "deleting: $lo_list..."
        foreach lofile $lo_list {
                file delete $lofile
        	}
        puts  $lofout "...Done "


        puts $ldbfout "$now [clock clicks] tar xmvzf $tgz_file ..."
        if {[catch {eval exec tar xmvzf $tgz_file} errmsg]} {
        	# il tar ritorna lo stato di errore 1 quando estrae un file con data futura
	        # il file viene estratto comunque senza problemi
	        # il messaggio di warning e':
	        # tar: L'archivio contiene l'orario futuro ...
	        # cerco 'futuro' nel messaggio per vedere se mi trovo in questa situazione di non-errore
                if {[string first futuro $errmsg]<0} {
                        puts $ldbfout "$now [clock clicks] $errmsg "
                        puts $lofout "$now [clock clicks] $errmsg "
                        return -code error "$errmsg"
                        }
                }



        # rimuovo il file.tgz
        puts $lofout "$now [clock clicks] deleting $tgz_file..."
        file delete $tgz_file
        set tgz_list [split [glob -nocomplain -- *lofiles.tgz] { }]
        puts $lofout "$now [clock clicks] tgz_list: $tgz_list"


	# importing lofiles (host_table_loid.lo) in pg_largeobject
	puts $lofout "$now [clock clicks] importing lofiles in  $pgrdVar($dbname,inbox)"
        set abslo_list [split [glob -nocomplain -directory $pgrdVar($dbname,inbox) -- *.lo] { }]


        foreach abslofile $abslo_list {

		# extract remote_loid rloid from abslofile (inbox dir)/siteN_table_rloid.lo
		set loname [file rootname [file tail $abslofile]] ;# -> siteN_table_rloid
		#regexp {.*_([0-9]+)$} $loname match rloid ;# -> rloid
		if {![regexp {site([0-9]+)_.*_([0-9]+)$}  $loname match rsid rloid]} {
			# impossible to define siteID and rloid
			puts $lofout "$now [clock clicks] $loname: it is impossible to define siteID and rloid!"
			set rsid 0
			set rloid 0
			} 

	        # importing lofile in pg_largeobject setting lobjId=loid
		# and populating pgr_lomap (LO mapping table)
        	pg_exec $lconn "BEGIN"
	        set lobjId [pg_lo_import $lconn $abslofile]
		pg_exec $lconn "insert into pgr_lomap (rsiteid, rloid, lloid) values ($rsid, $rloid, $lobjId)"
	        pg_exec $lconn "END"


	        };# end of  foreach abslofile...


        puts  $lofout "$now [clock clicks]...Done"

        # removing files.lo in  $pgrdVar($dbname,inbox)
        puts $lofout "$now [clock clicks] removing files.lo in  $pgrdVar($dbname,inbox)"
        set lo_list [split [glob -nocomplain -directory $pgrdVar($dbname,inbox) -- *.lo] { }]
        puts  $lofout "deleting: $lo_list"
        foreach lofile $lo_list {
                file delete $lofile
                }
        puts  $lofout "...Done "

	};# end of foreach tgz_file...


set now [clock format [clock second] -format %T]
puts $ldbfout "$now [clock clicks]  Large Object Import:   End of procedure"
puts $lofout "$now [clock clicks]  Large Object Import:   End of procedure"

};# end of proc



#/////////////////////////////////////////////////
#/////////////////////////////////////////////////

proc LO_update { dbname } {
global pgrdVar
set lfout $pgrdVar(pgrdlog)
set lhost $pgrdVar($dbname,lhost)
set ldbfout $pgrdVar($dbname,pgrdlog)
set lsite $pgrdVar($dbname,lsite)
set lconn $pgrdVar($dbname,lconn)
set lofout $pgrdVar($dbname,lofout)

set now [clock format [clock second] -format %T]
puts $ldbfout [format "$now [clock clicks] Large Object Update\t$dbname"]
puts $lofout [format "\n$now [clock clicks] Large Object Update\t$dbname"]

set now "        "

if {![info exists pgrdVar($dbname,lotables)]} {
        setLOpgrdVar $dbname
        }


# find lomap(siteid,rloid)=lloid
puts $ldbfout "$now [clock clicks] poputating lomap array..."
puts $lofout "$now [clock clicks] poputating lomap array..."
set query "select * from pgr_lomap"
catch {unset lomap}
pg_select $lconn $query row {
        set lomap($row(rsiteid),$row(rloid)) $row(lloid)
        }
puts $lofout "$now [clock clicks] ...Done"


# update table according to lomap array
puts $ldbfout "$now [clock clicks] update table according to lomap array..."
puts $lofout "$now [clock clicks] update table according to lomap array..."
foreach flowtable $pgrdVar($dbname,lotables) {
        # find tablename from flowtablename
        regexp {(.+)_flow} $flowtable match table

	# flowtable lofields
	set lofields  $pgrdVar($dbname,$flowtable,lofields)
        set lofieldsString [join $pgrdVar($dbname,$flowtable,lofields) ", "]

	set rec_state first_rec
	set lop ""
	catch {unset loidarray}
	foreach lofield $lofields {
		# buffer loid array 4 record matching
	        set loidarray($lofield) ""
	        }

	# find remote passed flows
        set query "select pgr_grid, pgr_gfid, pgr_sid_f(pgr_gfid) as siteid, pgr_optype, $lofieldsString from  [myquote $flowtable] where pgr_sid_f(pgr_gfid)!=$lsite  and pgr_flow_flag=\'Y\' order by pgr_grid asc, pgr_gfid asc"
        puts $lofout $query

	puts $lofout [format "\npgr_grid | pgr_optype | [join $lofields " | "] | rec_state | [join $lofields " | "]\n"]

	# analizing flowtables rows
	pg_select $lconn $query Trans {
	        if {[string compare $rec_state first_rec]!=0} {
        	        if {$grid == $Trans(pgr_grid)} {
                	        set rec_state same_rec
                        	} else {
	                        set rec_state new_rec
                	        }
	                }

	        switch -exact -- $rec_state {
        	        first_rec {
                	        set grid $Trans(pgr_grid)
                        	foreach lofield $lofields {
                                	set loidarray($lofield) $Trans($lofield)
	                                }
        	                set rec_state new_rec
                	        }
	                new_rec {
				# verify last loidarray 4 loid update
				set lofields2update {}
				set lofields2updateset {}
                	        foreach lofield $lofields {
                        	        if {[string compare $loidarray($lofield) ""]!=0 && [string compare $lop "D"]!=0} {
						lappend lofields2update $lofield 
						lappend lofields2updateset "$lofield = $lomap($lsiteid,$loidarray($lofield))::lo"
	                                        }
					}
		                if {[llength $lofields2update]>0} {
					# save old LO values in flowtable
					#(new inserted rows are not visible in this pg_select loop:Ok)
					set lofields2updString [join $lofields2update ", "]
					set rquery "insert into [myquote $flowtable] (pgr_gfid, pgr_optype, pgr_when, pgr_username, pgr_flow_flag, pgr_grid, $lofields2updString) select  pgr_gxid_f($lsite, nextval(\'pgr_fid_s\')), \'LO\', \'now\', \'updated_LO\', \'Y\', \'$grid\', $lofields2updString from $table where  pgr_grid = $grid"
					puts $lofout "$rquery"
					set ris [query_exec $dbname $rquery]
					puts $lofout "$ris"

					# update table
        		                set lofields2updsetString [join $lofields2updateset ", "]
                		        set uquery "update $table set $lofields2updsetString where pgr_grid = $grid"
                        		puts $lofout $uquery
		                        set res [pg_exec $lconn $uquery]
        		                set err [pg_result $res -error]
					set err [string trimright "$err"]
                		        pg_result $res -clear
		                        set lquery "insert into pgr_qryengine_log values ($lsiteid,\'[myquote $table]\',0,$grid,\'LO\',\'now()\',\'replicator\',\'$err\');"
        		                puts $lofout "$lquery"
                		        set ris [query_exec $dbname $lquery]
                        		puts $lofout "$ris"
		                        }
				# save actual loidarray
				foreach lofield $lofields {
	        	                set loidarray($lofield) $Trans($lofield)
					}
                        	}
	                same_rec {
        	                foreach lofield $lofields {
					# loidarray($lofield)=0 -->  no loid update required
                	                if {[string compare $loidarray($lofield) $Trans($lofield)]==0} {
                        	                set loidarray($lofield) ""
                                	        } else {
                                        	set loidarray($lofield) $Trans($lofield)
	                                        }
        	                        }
                	        }
	                default {
        	                puts $ldbfout "find_LO_list Error: $rec_state not in <first_rec | new_rec | same_rec>"
                	        return -1
                        	}
	                } ;# end of switch
	
	        	set loglist {}
		        lappend loglist $Trans(pgr_grid) $Trans(pgr_optype)
	        	foreach lofield $lofields { lappend loglist $Trans($lofield) }
		        lappend loglist $rec_state
        		foreach lofield $lofields { lappend loglist $loidarray($lofield) }
		        set logString [join $loglist " | "]
	
		        puts $lofout $logString

			# save op_type, siteid & grid 4 next record match
		        set lop $Trans(pgr_optype)
			set lsiteid $Trans(siteid)
			set grid $Trans(pgr_grid)

                } ;# end of pg_select

		# it was the last_rec:
		set lofields2update {}
		set lofields2updateset {}
		foreach lofield $lofields {
			if {[string compare $loidarray($lofield) ""]!=0 && [string compare $lop "D"]!=0} {
				lappend lofields2update $lofield
                                lappend lofields2updateset "$lofield = $lomap($lsiteid,$loidarray($lofield))::lo"
                                }
                        }
                if {[llength $lofields2update]>0} {

                        # save old LO values in flowtable
                        #(new inserted rows are not visible in this pg_select loop:Ok)
                        set lofields2updString [join $lofields2update ", "]
                        #set rquery "insert into [myquote $flowtable] (pgr_gfid, pgr_optype, pgr_when, pgr_username, pgr_flow_flag, pgr_grid, $lofields2updString) select  pgr_gxid_f($lsite, nextval(\'pgr_fid_s\')), \'LO\', \'now\', \'updated_LO\', \'Y\', \'$grid\', $lofields2updString from $table where  pgr_grid = $grid"
                        set rquery "insert into [myquote $flowtable] (pgr_gfid, pgr_optype, pgr_when, pgr_username, pgr_flow_flag, pgr_grid, $lofields2updString) select  0, \'LO\', \'now\', \'updated_LO\', \'Y\', \'$grid\', $lofields2updString from $table where  pgr_grid = $grid"
			puts $lofout "$rquery"
                        set ris [query_exec $dbname $rquery]
			puts $lofout "$ris"

                        # update table
                        set lofields2updsetString [join $lofields2updateset ", "]
                        set uquery "update $table set $lofields2updsetString where pgr_grid = $grid"
                        puts $lofout "$uquery"
                        set res [pg_exec $lconn $uquery]
                        set err [pg_result $res -error]
			set err [string trimright "$err"]
                        pg_result $res -clear
                        #set lquery "insert into pgr_qryengine_log values ($lsiteid,\'[myquote $table]\',0,$grid,\'LO\',\'now()\',\'replicator\',\'$err\');"
                        set lquery "insert into pgr_qryengine_log values ($lsiteid,\'[myquote $table]\',pgr_gxid_f($lsite, nextval(\'pgr_fid_s\')),$grid,\'LO\',\'now()\',\'replicator\',\'$err\');"
                        puts $lofout $lquery
                        set ris [query_exec $dbname $lquery]
                        puts $lofout "$ris"
                        }


        } ;# end of foreach flowtable...


# LO vacuming in pg_largeobject (transferred to RCLOSE)
#set now [clock format [clock second] -format %T]
#puts $lofout "$now [clock clicks]  LO vacuming in pg_largeobject"
#puts $ldbfout "$now [clock clicks]  LO vacuming in pg_largeobject"
#if [catch {LO_vacuum $dbname} err] {
#	set now [clock format [clock second] -format %T]
#        puts  "$now [clock clicks] WARNING: LO_vacuum $dbname ERR - $err"
#        puts $lofout "$now [clock clicks] WARNING: LO_vacuum $dbname ERR - $err"
#        puts $ldbfout "$now [clock clicks] WARNING: LO_vacuum $dbname ERR - $err"
#        }


set now [clock format [clock second] -format %T]
puts $lofout "$now [clock clicks]  Large Object Update:   End of procedure"
puts $ldbfout "$now [clock clicks]  Large Object Update:   End of procedure"


};# end of proc



#/////////////////////////////////////////////////
#/////////////////////////////////////////////////



proc LO_update_back { dbname } {
# rollback LO update operations made in  LO_update
# and logged in table_flow as 'updated_LO' user operations

global pgrdVar
set lfout $pgrdVar(pgrdlog)
set lhost $pgrdVar($dbname,lhost)
set ldbfout $pgrdVar($dbname,pgrdlog)
set lsite $pgrdVar($dbname,lsite)
set lconn $pgrdVar($dbname,lconn)
set lofout $pgrdVar($dbname,lofout)

set now [clock format [clock second] -format %T]
puts $ldbfout [format "$now [clock clicks] Large Object Update Rollback\t$dbname"]
puts $lofout [format "\n$now [clock clicks] Large Object Update Rollback\t$dbname"]

set now "        "

if {![info exists pgrdVar($dbname,lotables)]} {
        setLOpgrdVar $dbname
        }

# rollback table according to flowtable LO logs
puts $ldbfout "$now [clock clicks] rollback tables according to flowtable LO logs..."
puts $lofout "$now [clock clicks] rollback tables according to flowtable LO logs..."
foreach flowtable $pgrdVar($dbname,lotables) {
        # find tablename from flowtablename
        regexp {(.+)_flow} $flowtable match table
	puts $lofout "$now $table:"

        # flowtable lofields
        set lofields  $pgrdVar($dbname,$flowtable,lofields)
        set lofieldsString [join $pgrdVar($dbname,$flowtable,lofields) ", "]

	set flowquery "select pgr_grid, $lofieldsString from $flowtable where pgr_username=\'updated_LO\'"
	pg_select $lconn $query flowrow {

	        #set lofields2update {}
                set lofields2updateset {}
                foreach lofield $lofields {
                	if {[string compare $flowrow($lofield) ""]!=0} {
                        	#lappend lofields2update $lofield
                                lappend lofields2updateset "$lofield = $flowrow($lofield)::lo"
                                }
                        }

                if {[llength $lofields2update]>0} {
                        # update table
                        set lofields2updsetString [join $lofields2updateset ", "]
                        set uquery "update $table set $lofields2updsetString where pgr_grid = $flowrow(pgr_grid)"
                        puts $lofout "$uquery"
                        set res [pg_exec $lconn $uquery]
                        set err [pg_result $res -error]
                        set err [string trimright "$err"]
                        puts $lofout "$ris"
			};# end of if

		};# end of pg_select


	};# end of foreach



set now [clock format [clock second] -format %T]
puts $lofout "$now [clock clicks]  Large Object Update Rollback:   End of procedure"
puts $ldbfout "$now [clock clicks]  Large Object Update Rollback:   End of procedure"


};# end of proc



