#--------------------------------------------------------------------
# query engine: 
# remote accepted operations are submitted or
# local rejected operations are rolled back
# accordingly to flow_flags in auxiliary tables;
# <dbname>_rej.sql file is created in log directory 
# to eventually re-submit all 
# local rejected operations :)
#--------------------------------------------------------------------
# it is also used for rollback after replica errors:
# query_engine dbname tablename local replicator back
#--------------------------------------------------------------------


proc remap_dir { op } {
# return the same operation
# 	I   ->   I
#	Un  ->   Uo
#	Uo  ->   Un
#       D   ->   D

	switch $op {
		I	{ set op_remap I }
		Un	{ set op_remap U }
		Uo	{ set op_remap U }
		D	{ set op_remap D }
		}
	return $op_remap
	}


#------------------------------------------


proc remap_inv { op } {
# define inverse operations
#       I   ->   D
#	Un  ->   Uo
#	Uo  ->   Un
#       D   ->   I

        switch $op {
                I       { set op_remap D }
                Un      { set op_remap U }
                Uo      { set op_remap U }
                D       { set op_remap I }
                }
	return $op_remap
        }



#------------------------------------------


proc query_engine { dbname tablename trans_site trans_type direction {abssqlfname {}}} { 
global pgrdVar

set Version 1.09.3

set lconn $pgrdVar($dbname,lconn)

# this procedure log file 
set logfname [file join $pgrdVar($dbname,log) queryEngine.log]
set lfout [open $logfname a+] 


#if {[string compare $abssqlfname ""] == 0 } {
	# temp log_engine file (copy performance is better than multiple insert)
	#set pgrlogfname [file join $pgrdVar($dbname,log) pgr_qryengine.log]
	#set pgrlogh [open $pgrlogfname w+]
	#fconfigure $pgrlogh -buffering none
#}


set usage [format "\a\n\nUsage: \n   queryEngine  <dbname> <tablename> <local | remote | allsites> <passed | rejected | alltransactions> <forward | back>\n {absolute_path_of_sql_filename}" ]



# define the auxiliary table
set tabellaflow ${tablename}_flow

# define LO fields list for tabellaflow
if {[info exists pgrdVar($dbname,$tabellaflow,lofields)]} {
	set lofields  $pgrdVar($dbname,$tabellaflow,lofields)
	} else {
	set lofields {}
	}


# define the local site 
set locale $pgrdVar($dbname,lsite)

# now we can define the select query

# local, remote or all transactions: <local | remote | allsites>
switch -exact -- $trans_site {

	local {
		set fromsite_clause " and pgr_sid_f(pgr_gfid)=$locale"  ;       # local transactions
		}

	remote {
		set fromsite_clause " and pgr_sid_f(pgr_gfid)!=$locale"  ;      # remote transactions
		}

	allsites {
		set fromsite_clause " "  ;       # all transactions
		}

	default {
		puts $lfout "$trans_site not in <local | remote | allsites>"
		close $lfout
		return -1
		}
	} ;# end of switch



# what transactions: <passed | rejected | alltransactions | replicatorLO>
switch -exact -- $trans_type {

        passed {
                set trans_flag " and pgr_flow_flag=\'Y\'"
                }

        rejected {
                set trans_flag " and pgr_flow_flag!=\'Y\'"
                }

        alltransactions {
		set trans_flag " "
                }

	replicatorLO {
		set trans_flag " and pgr_username=\'replicator\'"
		}

        default {
                puts $lfout "$trans_type  not in <passed | rejected | alltransactions | replicatorLO>"
		close $lfout
                return -1
                } 
        } ;# end of switch



# goforward o comeback: <forward | back>
switch -exact -- $direction {

        forward {
	        set op "Uo" ; # riga da saltare tra Uo e Un
		set op_remap_proc remap_dir ;# procedura di rimappatura operazioni
	        set ordertype "asc" ;     # ordine  crescente 
                }

        back {
                set op "Un" ; # riga da saltare tra Uo e Un
		set op_remap_proc remap_inv ;# procedura di rimappatura operazioni
                set ordertype "desc" ;    # ordine  decrescente 
                }

        default {
                puts $lfout "$direction not in <forward | back>"
		close $lfout
                return -1
                }
        } ;# end of switch 







set now [clock format [clock seconds]]

puts $lfout [format "\nTransactions Engine version $Version - $now - \n $dbname $tablename $trans_site $trans_type $direction\n"]


#------------------------------------------------------

# create a list with <tablename> fields 
# selecting attname (with attnum>0) from  pg_attribute 

set quotchar {"}

set fields {} ;# inizializza la lista fields

# pg_select does not distinguish between NULL value and null string "",
# always returning "", so we have to extract this info from database
set notnullfields {}
set query "select attname, attnotnull from pg_attribute where attnum>0 and attrelid=(select oid from pg_class where relname=\'[myquote $tablename]\')"
pg_select $lconn $query  fieldname {
	#set fields [linsert [set fields] end $fieldname(attname)]; # aggiunge in coda alla lista il nuovo campo
	# aggiunge in coda alla lista il nuovo campo
	lappend fields $fieldname(attname)
	#lappend fields $fieldname(attname)
	# aggiornaa lista dei campi dichiarati NOT NULL
	if {[string compare  $fieldname(attnotnull) "t"] == 0} {
		lappend notnullfields  $fieldname(attname)
		}
	}

if {[llength $fields]<=0} {
	set err "$tablename does not exist in $dbname!"
	puts $lfout $err
	close $lfout
	return  -code error $err
	}







puts $lfout [format "working on   $tablename  \{ $fields \}\n"]


# transactions engine
# we read from tablename auxiliary table (tablename_flow)
# and reconstruct related users operations

#set query "select * from  [myquote $tabellaflow] where pgr_optype!=\'$op\' ${fromsite_clause} ${trans_flag} order by pgr_when  $ordertype, pgr_gfid  $ordertype"
set query "select  pgr_sid_f(pgr_gfid) as siteid, * from  ${quotchar}[myquote $tabellaflow]${quotchar} where pgr_optype!=\'$op\' and  pgr_optype!=\'LO\' ${fromsite_clause} ${trans_flag} order by pgr_when  $ordertype, pgr_gfid  $ordertype"

puts $lfout  [format "[clock clicks] query:\n$query"]

pg_select $lconn "$query" trans {

	set now [clock format [clock seconds]]

	set optype [$op_remap_proc $trans(pgr_optype)]

	switch -exact -- $optype {	

# INSERT
		I {
			# costruisco la query del tipo
			#	INSERT INTO class_name [(attr1, ...attrN)]
			#	VALUES (expr1,..exprN)

			# for remote operations we DO NOT skip LO fields inserting
                        # because it should be declared as NOT NULL ...


			set query "INSERT INTO ${quotchar}[myquote $tablename]${quotchar} (pgr_grid" 

			# fields2 e' la lista dei campi escludendo grid
			set rep_ind [lsearch -exact $fields pgr_grid]
			set fields2 [lreplace $fields $rep_ind $rep_ind]

			foreach fd $fields2 {
				append query ",${quotchar}[myquote $fd]${quotchar}"
				}
			append query ") VALUES ($trans(pgr_grid)"

                        foreach fd $fields2 {
				set fvalue [myquote $trans($fd)]
				if {[string compare $fvalue ""]!=0} {
                                	append query ",\'$fvalue\'"
					} else {
					if {[lsearch -exact $notnullfields $fd] != -1} {
						append query ",\'\'"
						} else {
						append query ",NULL"
						}
					}
                                }
			append query "); "

			} ;# end of Insert





# UPDATE
		U {
                        # fields2 e' la lista dei campi escludendo grid
			# che non deve essere aggiornato
                        set rep_ind [lsearch -exact $fields pgr_grid]
                        set fields2 [lreplace $fields $rep_ind $rep_ind]

			# for remote operations we skip LO fields updating
			# because they are efficiently done in LO procedures
			# Ex. if a LO field is not relly updated,  we do not export
			# LO file so we do not have to change local pg_largeobject.loid reference 
			if {[string compare $trans(siteid) $locale] != 0} {
				foreach field $lofields {
					set rep_ind [lsearch -exact $fields2 $field]
					set fields2 [lreplace $fields2 $rep_ind $rep_ind]
					}
				}

			set query "UPDATE ${quotchar}[myquote $tablename]${quotchar} SET "
                        foreach fd $fields2 {
				set fvalue [myquote $trans($fd)]
                                if {[string compare $fvalue ""]!=0} {
                                        append query "${quotchar}${fd}${quotchar}=\'$fvalue\',"
                                        } else {
					if {[lsearch -exact $notnullfields $fd] != -1} {
						append query "${quotchar}${fd}${quotchar}=\'\',"
						} else {
	                                        append query "${quotchar}${fd}${quotchar}=NULL,"
						}
                                        }
                                }

			set query [string trimright $query ","] ; # elimino la virgola finale
			append query " WHERE pgr_grid=\'$trans(pgr_grid)\';"

			} ;# end of Update



# DELETE
		D {
			set query "DELETE FROM ${quotchar}[myquote $tablename]${quotchar} WHERE pgr_grid=\'$trans(pgr_grid)\';" 
                        #append query_string $query

			} ;# end of Delete


		default {
			puts $lfout "ERROR-invalid operator $trans(pgr_optype)"
			puts $lfout "Table:		[myquote $tablename]"
			puts $lfout "Transaction_ID:    $trans(pgr_gfid)"
			set ris [query_exec $dbname "insert into pgr_qryengine_log values ($pgrdVar($dbname,lsite),\'[myquote $tablename]\',$trans(pgr_fromsite),$trans(pgr_floid),$trans(pgr_mastersite),$trans(pgr_boid),\'$trans(pgr_optype)\',\'$trans(pgr_when)\',\'$trans(pgr_username)\',\'ERROR-invalid queryEngine operator $trans(pgr_optype)\');"]
			} ;# end of default


		} ;# end of switch


	if {[string compare $abssqlfname ""] == 0 } {

		# we execute query, logging into pgr_qryengine_log table

		#set ris [query_exec $dbname "begin work;"]
		#puts $lfout "begin work: $ris"

	        #set dbhandle $pgrdVar($dbname,lconn)
	        set res [pg_exec $lconn $query]
	        set err [pg_result $res -error]
		# remove eventually newline at the end of the err string
		# (to avoid error during copy)
		set err [string trimright "$err"]
	        pg_result $res -clear

		# insert replicated op. info into pgr_qryengine_log
		set query "insert into pgr_qryengine_log values ($pgrdVar($dbname,lsite),\'[myquote $tablename]\',$trans(pgr_gfid),$trans(pgr_grid),\'$optype\',\'$trans(pgr_when)\',\'$trans(pgr_username)\',\'$err\');"
		#puts $lfout "$query..."
		set ris [query_exec $dbname $query]
		#puts $lfout "$ris"

	 	#set ris [query_exec $dbname "commit;"]
		#puts $lfout "commit: $ris"

		# log into log_engine file
		#puts $pgrlogh "$pgrdVar($dbname,lsite)|[myquote $tablename]|$trans(pgr_gfid)|$trans(pgr_grid)|$trans(pgr_optype)|$trans(pgr_when)|$trans(pgr_username)|$err"
		#puts $pgrlogh "$pgrdVar($dbname,lsite)|[myquote $tablename]|$trans(pgr_gfid)|$trans(pgr_grid)|$optype|$trans(pgr_when)|$trans(pgr_username)|$err"

		} else {	

		# query engine is used by save_rejected_operations procedure
		# to save local rejected operations in a sql file
		# identified by pgrdVar($dbname,qesql)
	
		puts $pgrdVar($dbname,qesql) $query
	
		}


	} ;# end of pg_select



if {[string compare $abssqlfname ""] == 0 } {

	puts $lfout "[clock clicks] table $tablename updated"

	#close $pgrlogh

	# copy log_engine file into pgr_qryengine_log table
	#set query "copy pgr_qryengine_log from \'$pgrlogfname\' using delimiters \'|\'"
	#puts $lfout "[clock clicks] $query..."
	#set ris [query_exec $dbname $query]
	#puts $lfout "[clock clicks] $ris"

	} else {

	puts $lfout "[clock clicks] file ${dbname}_rej.sql created!"

	}


set now [clock format [clock seconds]]
puts $lfout [format "\nTransactions Engine:   End of procedure on site $locale - $now -\n"]
close $lfout

}



