#-----------------------------------------------
#               RESOLVER
#-----------------------------------------------
# controlla e risolve eventuali conflitti nelle tabelle dei flussi, secondo gli
# algoritmi risolutivi specifici per ogni tabella (indicati in pgr_tables).
# Ogni transazione viene marchiata con 'Y' o 'N', in modo che il motore
# query_engine possa poi lavorarci sopra.
# (Di default il flag e' posto dai triggers a 'Y')


proc resolver { dbname  } {
global pgrdVar
set ldbfout $pgrdVar($dbname,pgrdlog)
set lhost $pgrdVar($dbname,lhost)

set quotchar {"}

set now [clock format [clock second] -format %T]
puts $ldbfout  "$now [clock clicks] $dbname Resolver "

set now "        "

# vacuuming database
#set query "vacuum "
#puts $ldbfout "$now [clock clicks] $query..."
#set ris [query_exec $dbname $query]
#puts $ldbfout "$now [clock clicks] ...$ris "


# creating auxiliary tables indexes 
foreach reptable $pgrdVar($dbname,$lhost,tables) {
        set flowtable ${reptable}_flow

	# vacuuming flowtable
	set query "vacuum [myquote ${quotchar}$flowtable${quotchar}]"
	puts $ldbfout "$now [clock clicks] $query..."
	set ris [query_exec $dbname $query]
	puts $ldbfout "$now [clock clicks] ...$ris "

        # create index ${flowtable}_idx
        set query "create index [myquote ${quotchar}${flowtable}_idx${quotchar}] on [myquote ${quotchar}${flowtable}${quotchar}] (pgr_grid, pgr_when, pgr_gfid)"
        puts $ldbfout  "$now [clock clicks] $query..."
        if [catch {query_exec $dbname $query} err] {
                puts $ldbfout "$now [clock clicks] ...$err"
                } else {
                puts $ldbfout "$now [clock clicks] ... Done"
                }
        #set ris [query_exec $dbname $query]
        #puts $ldbfout  "$now [clock clicks] ...$ris"
	}

# for each auxiliary table to resolv on this local site we select the conflict resolution algorithm from pgr_tables
pg_select $pgrdVar($dbname,lconn) "select pgr_sid, tablename, resolv_site, resolver from  pgr_lsite A, pgr_tables B where B.resolv_site = A.pgr_sid" localtable {

        set table $localtable(tablename)
        set res_meth $localtable(resolver)

        switch -exact -- $res_meth {


                NO      {       ;# No conflict resolution algorithm to apply

                        puts $ldbfout [format "$now [clock clicks] \tAny conflict resolution algorithm to apply on $table"]

                        }



                default {  ;# external resolvers res_meth.tcl

                        set res_file [file join $pgrdVar(pgrd_resolv) ${res_meth}.tcl]
                        if [file exists $res_file] {
				set dirlog $pgrdVar($dbname,log)
                                puts $ldbfout "$now [clock clicks] exec $res_file $dbname $table $dirlog ..."
                                if {[catch {eval exec  $res_file $dbname $table $dirlog} errmsg]} {
					puts $ldbfout "$now [clock clicks] $table -  $res_meth ...Errore in fase di esecuzione algoritmo"
					return -code error "$table - $res_meth ...Error: $errmsg"
					} else {
                                	puts $ldbfout "$now [clock clicks] $table - $res_meth ...done"
					}
                                } else {
                                puts $ldbfout "$now [clock clicks] $table - $res_meth ...Error:  algoritmo non riconosciuto!"
				return -code error "$table - $res_meth ... no recognized algorithm!"
                                }

                        }

                }


        } ;# end of pg_select

puts $ldbfout  "$now [clock clicks] $dbname End of Resolver "

} ;# end of proc resolver

#-----------------------------------------------------------




