#!/bin/sh
# the next line restart using pgtclsh \
exec pgtclsh "$0" "$@"


# One Site Only  conflict resolution algorithm

# Usage: 1SO.tcl <dbname> <tablename> <log_dir>

# if operations on <tablename> comes from 2 or more sites,
# this algorithm rejects all the operations



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
proc how_many_sites {flowtable} {
        global lconn
        set num 0
        set sites [list ]

        pg_select $lconn "select distinct pgr_sid_f(pgr_gfid) as fromsite from $flowtable"  Sites {
                incr num
                set sites [linsert [set sites] end $Sites(fromsite)]
                }

        # returns a list like {num_of_sites site1 site2 ... siteN}
        return [linsert [set sites] 0 $num]

        }
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
#		M A I N
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - -

global lconn


if { ($argc!=3) }  {
        puts stdout [format "\a\n\nUsage: \n1SO.tcl  <dbname> <tablename> <log_dir>\n"]
        return -1
        } 
set dbname [lindex $argv 0]
set tablename  [lindex $argv 1]
set logdir [lindex $argv 2]
set flowtable  ${tablename}_flow


set lfout [file join $logdir ${tablename}_1SO.log]
set hlfout [open $lfout w+]

# connecting to the database
puts  $hlfout [format "[clock clicks] locally connecting to database $dbname\n"]
if [catch { pg_connect $dbname  -port 5432} lconn ] {
        puts  $hlfout $lconn
        return -1
        }

puts $hlfout "1SO resolution algorithm applied on $flowtable in $dbname"

set now [clock format [clock second] -format %T]
puts $hlfout "$now [clock clicks] 1SO algorithm start..."

set sites [how_many_sites $flowtable]
set num_of_sites [lindex $sites 0]

if { $num_of_sites>1 } {

	set now [clock format [clock second] -format %T]
	set query "update $flowtable set pgr_flow_flag=\'N\'"
	puts $hlfout "$now [clock clicks] $query"
	set res [pg_exec $lconn $query]
	pg_result $res -clear 
        set msg "transactions on $tablename rejected because coming from too many sites: [lrange $sites 1 end]"


	set query "vacuum $flowtable"
	puts $hlfout "$now [clock clicks] $query"
	set res [pg_exec $lconn $query]
	pg_result $res -clear
        } else {
	set msg "transactions on $tablename coming from [lindex $sites 1] accepted"
	}



set now [clock format [clock second] -format %T]
puts  $hlfout "$now [clock clicks] One Site Only:  $msg"


# database disconnection
pg_disconnect $lconn


puts  $hlfout "Done"
close $hlfout



