#!/bin/sh
# the next line restart using pgtclsh \
exec pgtclsh "$0" "$@"


# First Site Only  conflict resolution algorithm

# Usage: FSO.tcl <dbname> <tablename> <log_dir>

# FSO scans the auxiliary table <tablename_flow> accepting only
# operations arised on the site wich has submitted the first transaction



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
proc first_site {flowtable} {
# ritorna il primo sito che ha fatto transazioni in un'epoca.
        global lconn 

        set res [pg_exec $lconn  "select min(pgr_sid_f(pgr_gfid))  from $flowtable where pgr_when=(select min(pgr_when) from $flowtable)"]

        pg_result $res -assign firstsite ; # firstsite <--  query result
	pg_result $res -clear
	return $firstsite([lindex [array names firstsite] 0])
        }


#- - - - - - - - - - - - - - - - - - - - - - - - - - - - -
#               M A I N
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - -

global lconn 

if { ($argc!=3) }  {
        puts stdout [format "\a\n\nUsage: \nFSO.tcl  <dbname> <tablename> <log_dir>\n"]
        return -1
        }
set dbname [lindex $argv 0]
set tablename [lindex $argv 1]
set logdir [lindex $argv 2]
set flowtable  ${tablename}_flow

set lfout [file join $logdir ${tablename}_FSO.log]
set hlfout [open $lfout w+]

puts $hlfout "FSO resolution algorithm applied on $flowtable in $dbname"

# connecting to database dbname
puts   $hlfout [format "[clock clicks] locally connecting to database $dbname\n"]
if [catch { pg_connect $dbname  -port 5432} lconn ] {
        puts  $hlfout $lconn
        return -1
        } 


set now [clock format [clock second] -format %T]
puts $hlfout "$now [clock clicks] FSO algorithm start..."

set firstsite [first_site $flowtable]



set now [clock format [clock second] -format %T]

set query "update $flowtable set pgr_flow_flag=\'N\' where  pgr_sid_f(pgr_gfid)!=$firstsite"
puts $hlfout "$now [clock clicks] $query..."
set res [pg_exec $lconn $query]
set ris [pg_result $res -status]
pg_result $res -clear
set now [clock format [clock second] -format %T]
puts  $hlfout "$now [clock clicks]...$ris"

set query "vacuum $flowtable"
puts $hlfout "$now [clock clicks] $query..."
set res [pg_exec $lconn $query]
set ris [pg_result $res -status]
pg_result $res -clear
set now [clock format [clock second] -format %T]
puts  $hlfout "$now [clock clicks]...$ris"

set msg "done site: $firstsite"
set msg2 "transactions rejected from other sites"
puts  $hlfout [format "\nFirst Site Only...%s\n\t%s" $msg $msg2]


# disconnection
pg_disconnect $lconn 


# close log file
close $hlfout



