#!/bin/sh
# the next line restart using pgtclsh \
exec pgtclsh "$0" "$@"


# Time Stamp  conflict resolution algorithm

# Usage: TS.tcl <dbname> <tablename> <log_dir>


# for each record,
# we find the site wich submitted the last operation and accept 
# only its operations


if { ($argc!=3) }  {
        puts stdout [format "\a\n\nUsage: \nTS.tcl  <dbname> <tablename> <log_dir>\n"]
        return -1
        }
set dbname [lindex $argv 0]
set tablename [lindex $argv 1] 
set logdir [lindex $argv 2]
set flowtable  ${tablename}_flow


set lfout [file join $logdir ${tablename}_TS.log ]
set hlfout [open $lfout w+]

# connecting to the database
puts  $hlfout [format "[clock clicks] locally connecting to database $dbname\n"]
if [catch { pg_connect $dbname  -port 5432} lconn ] {
        puts  $hlfout $lconn
        return -1
        }

puts $hlfout "TS resolution algorithm applied on $flowtable in $dbname"


# we define a dummy table rej_flowtable
# using a 'select into' and extracting 0 records
set query "select pgr_gfid into rej_$flowtable from $flowtable where pgr_gfid<0"
puts $hlfout $query
set res [pg_exec $lconn $query]
set ris [pg_result $res -error]
puts $hlfout $ris
pg_result $res -clear


set order desc
set rec_state init

set now [clock format [clock second] -format %T]
puts $hlfout "$now [clock clicks] TS algorithm start..."


set query "select pgr_gfid,pgr_sid_f(pgr_gfid) as sid,pgr_grid,pgr_when from  $flowtable where pgr_optype !='Uo' order by pgr_grid, pgr_when $order, pgr_gfid $order"
puts $hlfout $query

puts $hlfout "list of operatinons:"
puts $hlfout "pgr_gfid pgr_grid rec_state pgr_flow_flag"



pg_select $lconn $query  Trans {

	if {[string compare $rec_state init]==0} {
		set rec_state new_rec
		set first_site $Trans(sid)
                #set mastersite $Trans(pgr_mastersite)
                #set boid $Trans(pgr_boid)
		set grid $Trans(pgr_grid)

                } elseif {$grid == $Trans(pgr_grid)} {
                	set rec_state same_rec
                        } else {
                        set rec_state new_rec
                        set first_site $Trans(sid)
                        #set mastersite $Trans(pgr_mastersite)
                        #set boid $Trans(pgr_boid)
			set grid $Trans(pgr_grid)

                        }

	if {[string compare $rec_state new_rec]==0} {
		set Trans(pgr_flow_flag) "Y"
                } elseif {$first_site==$Trans(sid)} {
			set Trans(pgr_flow_flag) "Y"
                        } else {
			set Trans(pgr_flow_flag) "N"
                        }
	puts $hlfout "$Trans(pgr_gfid) $Trans(pgr_grid) $rec_state $Trans(pgr_flow_flag)"

	# insert values into rej_$flowtable
	if {[string compare $Trans(pgr_flow_flag) "N"]==0} {
		set query "insert into rej_$flowtable values ($Trans(pgr_gfid))"
		puts $hlfout $query
		set res [pg_exec $lconn $query]
		set ris [pg_result $res -error]
		puts $hlfout $ris
		pg_result $res -clear
		#puts $hlfout "$Trans(pgr_gfid)"
		} ;# end of if

        }       ;# end of pg_select


set now [clock format [clock second] -format %T]
puts $hlfout "$now [clock clicks] start updating $flowtable..."


# we update attribute flow_flags in flowtable
set query "update $flowtable set pgr_flow_flag = 'N' where pgr_gfid in (select * from rej_$flowtable)"
puts $hlfout $query
set res [pg_exec $lconn $query]
set ris [pg_result $res -error]
puts $hlfout $ris
pg_result $res -clear

set now [clock format [clock second] -format %T]
puts $hlfout "$now [clock clicks] end"

# dropping the dummy table
puts $hlfout "dropping the dummy table"
set res [pg_exec $lconn "drop table rej_$flowtable"]
set ris [pg_result $res -error]
puts $hlfout $ris
pg_result $res -clear


# database disconnection
pg_disconnect $lconn


puts  $hlfout "Done"
close $hlfout

