#!/bin/sh
# the next line restart using pgtclsh \
exec pgtclsh "$0" "$@"


# Master Priority Control  post_resolution algorithm

# Usage: mpc.tcl <dbname> <tablename> <log_dir>

# you can use this post_resolution algorithm 
# with record oriented conflict resolution algorithms 
# such as TS FTS FTO, obtaining a xxx-MP conflict resolution algorithm

# MP scans the auxiliary table <tablename_flow> and
# checks if operations accepted by xxx are originated on the mastersite:
# if xxx has rejected operations originated on the mastersite, MP accepts them
# and reject the others




if { ($argc!=3) }  {
        puts stdout [format "\a\n\nUsage: \nmpc.tcl  <dbname> <tablename> <log_dir>\n"]
	return -1
        }
set dbname [lindex $argv 0]
set tablename  [lindex $argv 1]
set logdir [lindex $argv 2]
set flowtable  ${tablename}_flow

set lfout [file join $logdir ${tablename}_mpc.log]
set hlfout [open $lfout w+]

# connecting to the database
puts $hlfout  [format "[clock clicks] locally connecting to database $dbname\n"]
if [catch { pg_connect $dbname  -port 5432} lconn ] {
        puts $hlfout $lconn
        return -1
        }


set now [clock format [clock second] -format %T]
puts $hlfout "$now [clock clicks] Master Priority Control algorithm start..."

pg_select $lconn "select oid, pgr_gfid, pgr_grid from $flowtable where pgr_sid_f(pgr_grid)=pgr_sid_f(pgr_gfid) and pgr_flow_flag=\'N\'" Trans {

	set query "update $flowtable set pgr_flow_flag=\'N\' where pgr_grid=$Trans(pgr_grid) and pgr_sid_f(pgr_gfid)!=pgr_sid_f($Trans(pgr_gfid)) and pgr_flow_flag=\'Y\'"
	puts $hlfout  "$now [clock clicks] $query..."
	set res [pg_exec $lconn $query]
	pg_result $res -clear

	set now [clock format [clock second] -format %T]
        set query "update $flowtable set pgr_flow_flag=\'Y\' where oid=\'$Trans(oid)\'"
	puts $hlfout  "$now [clock clicks] $query..."
        set res [pg_exec $lconn $query]
        pg_result $res -clear

        } 


set now [clock format [clock second] -format %T]
set query "vacuum $flowtable"
puts $hlfout  "$now [clock clicks] $query..."
set res [pg_exec $lconn $query]
pg_result $res -clear

# database disconnection
pg_disconnect $lconn

# close the log file
puts $hlfout  "Done"
close $hlfout



