CREATE FUNCTION pgr_delete() RETURNS opaque AS '
# delete trigger function 
spi_exec "SELECT getpgusername() AS connected_user"
if {[string compare $connected_user replicator]!=0} {
        # trigger fires when user is not replicator
	set quotchar {"}
        set lsid -1
        spi_exec "SELECT relname FROM pg_class WHERE oid=$TG_relid"
        spi_exec "SELECT pgr_sid AS lsid FROM pgr_lsite"
        spi_exec "SELECT permissions, monitor, locked FROM pgr_tables WHERE tablename=\'[quote $relname]\'"
        # returns if tablename is locked
        if {$locked} {
                elog NOTICE "$relname locked: no operations allowed"
                return SKIP
                }
        # data ownership model control
        # ROT: Read Only table
        # RWT: Read Write Table
        # RWLP: Read Write Local Partition only
        if {[string compare $permissions ROT]==0} {
                elog NOTICE "$relname is Read-Only: writing access denied"
                return SKIP
                }
	spi_exec "select pgr_sid_f($OLD(pgr_grid)) as sid"
        if {($sid!=$lsid) && ([string compare $permissions RWLP]==0)} {
                elog NOTICE "$relname Remote Partitions are Read-Only (RWLP data ownership model)"
                return SKIP
                } else {
                # selected record can be deleted
                # TG_relatts on relname: {} field1 field2 ...
                set field_list [lrange [set TG_relatts] 1 end]
                set query "INSERT INTO ${quotchar}${relname}_flow${quotchar} ("
                foreach field ${field_list} {
                        append query ${quotchar}[quote $field]${quotchar},
                        }
                append query "pgr_gfid,pgr_optype,pgr_when,pgr_username,pgr_flow_flag) VALUES ("
                foreach field ${field_list} {
                        if {[info exists OLD(${field})]} {
                                append query \'[quote $OLD(${field})]\',
                                } else {
                                append query NULL,
                                }
                        }
		spi_exec "select pgr_gxid_f ( $lsid, nextval(\'pgr_fid_s\') ) as gfid"
                append query \'$gfid\',\'D\',\'now\',\'\$connected_user\',\'Y\')
                spi_exec "$query"
                # updating pgr_monitor
                if {$monitor} {
                        spi_exec "UPDATE pgr_monitor SET d_acc = 1 + d_acc, when_occurred=\'now\' WHERE tablename=\'[quote $relname]\'"
                        }
                }
        return OK
        } else {
        # trigger does not fire when user is replicator
        return OK
        }
' LANGUAGE 'pltcl';
