CREATE FUNCTION pgr_insert() RETURNS opaque AS '
# insert trigger function 
spi_exec "SELECT getpgusername() as connected_user"
if {[string compare $connected_user replicator]!=0} {
        # trigger fires when user is not replicator
	set quotchar {"}
        set lsid -1
        spi_exec "SELECT relname FROM pg_class WHERE oid=$TG_relid"
        spi_exec "SELECT pgr_sid AS lsid FROM pgr_lsite"
        spi_exec "SELECT permissions, monitor, locked FROM pgr_tables WHERE tablename=\'[quote $relname]\'"
        # returns if tablename is locked
        if {$locked} {
                elog NOTICE "$relname locked: any operation allowed"
                return SKIP
                }
        # data ownership model control
        # ROT: Read Only Table
        # RWT: Read Write Table
        # RWLP: Read Write Local Partition only
        if {[string compare $permissions ROT]==0} {
                elog NOTICE "$relname is Read-Only: writing access denied"
                return SKIP
                }
        # a new record can be inserted
        spi_exec "select pgr_gxid_f($lsid, nextval(\'pgr_rid_s\')) as grid"
        set NEW(pgr_grid) $grid
        # updating flow-table
        # TG_relatts: {} field1 field2 ...
        set field_list [lrange [set TG_relatts] 1 end]
        set query "INSERT INTO ${quotchar}${relname}_flow${quotchar} ("
        foreach field ${field_list} {
                append query ${quotchar}[quote $field]${quotchar},
                }
        append query "pgr_gfid,pgr_optype,pgr_when,pgr_username,pgr_flow_flag) VALUES ("
        foreach field ${field_list} {
                if {[info exists NEW(${field})]} {
                        append query \'[quote $NEW(${field})]\',
                        } else {
                        append query NULL,
                        }
                }
        spi_exec "select pgr_gxid_f ( $lsid, nextval(\'pgr_fid_s\') ) as gfid"
        append query \'$gfid\',\'I\',\'now\',\'\$connected_user\',\'Y\')
        spi_exec "$query"
        # updating pgr_monitor
        if {$monitor} {
                spi_exec "UPDATE pgr_monitor SET i_acc = 1 + i_acc, when_occurred=\'now\' WHERE tablename=\'[quote $relname]\'"
                }
        return [array get NEW]
        } else {
        # trigger does not fire when user is replicator
        return OK
        }
' LANGUAGE 'pltcl';
