CREATE FUNCTION pgr_update() RETURNS opaque AS '
# update trigger function 
spi_exec "SELECT getpgusername() AS connected_user"
if {[string compare $connected_user replicator]!=0} {
        # trigger fires when user is not replicator
	set quotchar {"}
        set lsid -1
        spi_exec "SELECT relname FROM pg_class WHERE oid=$TG_relid"
        spi_exec "SELECT pgr_sid AS lsid FROM pgr_lsite"
        spi_exec "SELECT permissions, monitor, locked FROM pgr_tables WHERE tablename=\'[quote $relname]\'"
        # returns if tablename is locked
        if {$locked} {
                elog NOTICE "$relname locked: any operation allowed"
                return SKIP
                }
        # data ownership model control
        # ROT: Read Only Table
        # RWT: Read Write Table
        # RWLP: Read Write Local Partition only
        if {[string compare $permissions ROT]==0} {
                elog NOTICE "$relname is Read-Only: writing access denied"
                return SKIP
                }
	spi_exec "select pgr_sid_f($OLD(pgr_grid)) as sid"
        if {($sid!=$lsid) && ([string compare $permissions RWLP]==0)} {
                elog NOTICE "$relname Remote Partitions are Read-Only (RWLP data ownership model)"
                return SKIP
                } else {
                # selected record can be updated
                # prevents updating of pgr_grid 
                set NEW(pgr_grid) $OLD(pgr_grid)
                # we have to update the flow-table only
                # TG_relatts on relname: {} field1 field2 ...
                set field_list [lrange [set TG_relatts] 1 end]
                # dquery: direct transaction entry
                # iquery: inverse transaction entry
                set dquery "INSERT INTO ${quotchar}${relname}_flow${quotchar} ("
                foreach field ${field_list} {
                        append dquery ${quotchar}[quote $field]${quotchar},
                        }
                append dquery "pgr_gfid,pgr_optype,pgr_when,pgr_username,pgr_flow_flag) VALUES ("
                set iquery $dquery
                foreach field ${field_list} {
                        if {[info exists OLD(${field})]} {
                                set old \'[quote $OLD(${field})]\',
                                } else {
                                set old NULL,
                                }
                        append iquery $old
                        if {[info exists NEW(${field})]} {
                                append dquery \'[quote $NEW(${field})]\',
                                } else {
                                append dquery $old
                                }
                        }
		spi_exec "select pgr_gxid_f ( $lsid, nextval(\'pgr_fid_s\') ) as gfid"
                append dquery \'$gfid\',\'Un\',\'now\',\'\$connected_user\',\'Y\')
                append iquery \'$gfid\',\'Uo\',\'now\',\'\$connected_user\',\'Y\')
                spi_exec "$iquery"
                spi_exec "$dquery"
                # updating pgr_monitor
                if {$monitor} {
                        spi_exec "UPDATE pgr_monitor SET u_acc = 1 + u_acc, when_occurred=\'now\' WHERE tablename=\'[quote $relname]\'"
                        }
                return [array get NEW]
                }
        } else {
        # trigger does not fire when user is replicator
        return OK
        }
' LANGUAGE 'pltcl';
