#!/bin/sh
# the next line restart using pgtclsh \
exec pgtclsh "$0" "$@"


# Usage: doSlave <dbname>

# Update pgr_tables for <dbname> setting permissions to ROT
# so all replicated tables are Read Only


proc init {} {
global pgrdVar 

        if {[catch {read_pgr.conf} msg]} {
                #puts "[clock clicks]  $msg"
                return -code error $msg
                }

        if {[catch {source [file join $pgrdVar(pgrd_HOME) setup.tcl]} msg]} {
                #puts "[clock clicks]  $msg"
                return -code error $msg
                }

        if {[catch {setup} msg]} {
                #puts stderr $msg
                return -code error $msg
                }

        # Loading all defined procedures modules
        foreach module { common pgproc } {
                source [file join $pgrdVar(pgrd_proc) $module.tcl]
                }
        }


proc read_pgr.conf {} {
global pgrdVar 
        if {[catch {open /etc/pgr.conf r } conf]} {
                #puts "[clock clicks] $conf"
                #puts stderr $conf
                return -code error $conf
                }

        while {![ eof $conf]} {
                gets $conf line
                #puts $line
                if {[string first # $line]!=0} {
                        if {[regexp {^([^= ]+)=([^ ]+)} $line match var value]} {
                                #puts "match: $match"
                                #puts "$var=$value"
                                set pgrdVar($var) $value
                                }
                        }
                }


        if {![info exists pgrdVar(PGRBIN)]} {
                puts stderr "No  PGRBIN  variable in /etc/pgr.conf"
                puts stderr "Is ReplicaTools properly installed ?"
                return -code error "No  PGRBIN  variable in /etc/pgr.conf"
                }

        set pgrdVar(pgrd_HOME) $pgrdVar(PGRBIN)

        }



#///////////
# MAIN
#///////////

if {[catch {init} errmsg ]} {
        puts stderr $errmsg
        exit
        }


puts [format "\npgReplicator Version: $pgrdVar(version)\ndoSlave utility"]

# only replicator can run this script
set uname [ exec id -un]
if {[string compare $uname replicator]!=0} {
        puts [format "Only user replicator can run this script!\n\a"]
        exit
        }


if  [expr (($argc < 1) || ($argc > 1))] {
	puts [format "Usage: doSlave.tcl <dbname>\n\a"]
        exit
        }

#foreach elem [array names pgrdVar] { puts "$elem $pgrdVar($elem)"}

set dbname [lindex $argv 0]
set quotchar {"}

# connect to dbname
if [catch {db_connect $dbname localhost $pgrdVar(pg_port)} err] {
	puts [format "Error connecting to $dbname:\n$err"]
	return -1
	} else {
	set lconn $pgrdVar($dbname,lconn)
	}



# check if dbname is a replicated database
set isreplicated 0
set query "select count(*) as isreplicated from pg_class where relname = \'pgr_tables\'"
set qry_handle [query_exec $dbname $query -return_handle]
pg_result $qry_handle -assign risp
set isreplicated $risp(0,isreplicated)
pg_result $qry_handle -clear
if { $isreplicated == 0} {
        puts [format "\n$dbname is not a replicated database!\n"]
	db_disconnect $dbname
	return -1
	}



# update pgr_tables
set query "update pgr_tables set permissions = \'ROT\'"
puts "$dbname: $query"
set ris [query_exec $dbname $query]
puts $ris



# disconnect from dbname
db_disconnect $dbname

puts " "

