#!/bin/sh
# the next line restart using pgtclsh \
exec pgtclsh "$0" "$@"


# Usage: update_triggers <dbname>

# Update replica triggers for <dbname>


proc init {} {
global pgrdVar 

        if {[catch {read_pgr.conf} msg]} {
                #puts "[clock clicks]  $msg"
                return -code error $msg
                }

        if {[catch {source [file join $pgrdVar(pgrd_HOME) setup.tcl]} msg]} {
                #puts "[clock clicks]  $msg"
                return -code error $msg
                }

        if {[catch {setup} msg]} {
                #puts stderr $msg
                return -code error $msg
                }

        # Loading all defined procedures modules
        foreach module { common pgproc } {
                source [file join $pgrdVar(pgrd_proc) $module.tcl]
                }
        }


proc read_pgr.conf {} {
global pgrdVar 
        if {[catch {open /etc/pgr.conf r } conf]} {
                #puts "[clock clicks] $conf"
                #puts stderr $conf
                return -code error $conf
                }

        while {![ eof $conf]} {
                gets $conf line
                #puts $line
                if {[string first # $line]!=0} {
                        if {[regexp {^([^= ]+)=([^ ]+)} $line match var value]} {
                                #puts "match: $match"
                                #puts "$var=$value"
                                set pgrdVar($var) $value
                                }
                        }
                }


        if {![info exists pgrdVar(PGRBIN)]} {
                puts stderr "No  PGRBIN  variable in /etc/pgr.conf"
                puts stderr "Is ReplicaTools properly installed ?"
                return -code error "No  PGRBIN  variable in /etc/pgr.conf"
                }

        set pgrdVar(pgrd_HOME) $pgrdVar(PGRBIN)

        }



#///////////
# MAIN
#///////////

if {[catch {init} errmsg ]} {
        puts stderr $errmsg
        exit
        }

#foreach elem [array names pgrdVar] { puts "$elem $pgrdVar($elem)"}
#exit

puts [format "\npgReplicator Version: $pgrdVar(version)\nupdate triggers utility"]

# only replicator can run this script
set uname [ exec id -un]
if {[string compare $uname replicator]!=0} {
        puts "Only user replicator can run this script!"
        exit
        }


if  [expr (($argc < 1) || ($argc > 2))] {
	puts [format "\nUsage: update_triggers <dbname>\n\a"]
        exit
        }


set dbname [lindex $argv 0]
set quotchar {"}

# connect to dbname
#db_connect $dbname localhost $pgrdVar(pg_port)
#set lconn $pgrdVar($dbname,lconn)


#------------->>>
# connect to dbname
if [catch {db_connect $dbname localhost $pgrdVar(pg_port)} err] {
        puts [format "Error connecting to $dbname:\n$err"]
        return -1
        } else {
        set lconn $pgrdVar($dbname,lconn)
        }



# check if dbname is a replicated database
set isreplicated 0
set query "select count(*) as isreplicated from pg_class where relname = \'pgr_tables\'"
set qry_handle [query_exec $dbname $query -return_handle]
pg_result $qry_handle -assign risp
set isreplicated $risp(0,isreplicated)
pg_result $qry_handle -clear
if { $isreplicated == 0} {
        puts [format "\n$dbname is not a replicated database!\n"]
	db_disconnect $dbname
        return -1
        }


#-------------<<<



# check if all auxiliary tables are empty
# otherwise problems with triggers will arise.
set flowtables {}
set query "select tablename from pgr_tables"
pg_select $lconn $query tb {
	lappend  flowtables ${tb(tablename)}_flow
}

set noupdate 0
foreach flowtable $flowtables {
	set query "select count(*) as flot from ${quotchar}[myquote $flowtable]${quotchar}"
	set qry_handle [query_exec $dbname $query -return_handle]
	pg_result $qry_handle -assign risp
	set acc $risp(0,flot)
	pg_result $qry_handle -clear
	if {$acc>0} {
		incr noupdate
		puts "$flowtable: $acc tuples - No Trigger Update!"
		} else {
		puts "$flowtable: $acc tuples"
	}
}

if {$noupdate>0} {
	# triggers update is not possible!
	puts " "
	puts "Trigger update is not possible"
	puts "because some auxiliary table is not empty!"
	puts "Sync first $dbname, then update triggers"
	puts " "
	
} else {

	puts " "
	puts "   dropping old triggers and trigger functions in $dbname"
	catch {query_exec $dbname "drop table pgr_tables_temp;"}
	query_exec $dbname "select * into pgr_tables_temp from pgr_tables;"
	query_exec $dbname "delete from  pgr_tables;"
	catch {query_exec $dbname "drop function pgr_insert();"}
	catch {query_exec $dbname "drop function pgr_update();"}
	catch {query_exec $dbname "drop function pgr_delete();"}
	catch {query_exec $dbname "drop trigger pgr_autotrigger_i on pgr_tables;"}
	catch {query_exec $dbname "drop trigger pgr_autotrigger_u on pgr_tables;"}
	catch {query_exec $dbname "drop trigger pgr_autotrigger_d on pgr_tables;"}
	catch {query_exec $dbname "drop function pgr_auto_i();"}
	catch {query_exec $dbname "drop function pgr_auto_u();"}
	catch {query_exec $dbname "drop function pgr_auto_d();"}

	puts "   Creating new $dbname trigger functions"
	eval exec psql  $dbname -f  $pgrdVar(PGRBIN)/triggers/i_func.sql 
	eval exec psql  $dbname -f  $pgrdVar(PGRBIN)/triggers/u_func.sql 
	eval exec psql  $dbname -f  $pgrdVar(PGRBIN)/triggers/d_func.sql 

	eval exec psql  $dbname -f $pgrdVar(PGRBIN)/triggers/create_triggers.sql 

	query_exec $dbname "insert into pgr_tables select * from pgr_tables_temp;"
	catch {query_exec $dbname "drop table pgr_tables_temp;"}

	puts "Done!"
	puts " "
}


# disconnect from dbname
db_disconnect $dbname


